/*
 * ============================================================================
 * Licensed Materials - Property of IBM
 * Project  Zero
 *
 * (C) Copyright IBM Corp. 2007  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 * ============================================================================
 */

// 
// File: /app/resources/applications.groovy
// 
// This script provides a limited REST API for creating new applications. Once 
// an application has been created, it is represented by a directory on the 
// virtual file system (see files.groovy).
//

import zero.ide.services.CliTaskInvoker
import zero.json.Json
import zero.ide.services.AppUtils
import zero.ide.services.FileUtils
import zero.ide.services.ModuleGroupTaskInvoker
import zero.management.appconfig.Application
import zero.util.XMLEncoder
import zero.util.UTF8FileWriter;

import zero.appbuilder.resources.MessageHandler

import javax.xml.parsers.ParserConfigurationException
import org.xml.sax.SAXParseException

class FileExclusion {
  public static String[] exclusions;
}


/**
 *
 * @success 200 Returns list of applications known by the Application Builder.
 * @format application/json
 * @example
 * [
 *   {
 *     "id": "1",
 *     "name": "my.app",
 *     "path": "path/to/parent"
 *   },
 *   {
 *     "id": "2",
 *     "name": "other.app",
 *     "path": "path/to/other/parent"
 *   },
 *    ...
 * }
 *
 */
def onList()
{
	def apps = []
	def keysToDelete = []

	// Walk the stored values first; delete 'missing' entries later
	// (avoid ConcurrentModificationExceptions)
	zget('/storage/applications', [:]).each { entry ->
		def appid = entry.key
		def apphome = AppUtils.getApphome(appid)
		if (apphome) {
			apps += [id:entry.key, name:entry.value.name, path:entry.value.path] + getAppDetails(apphome)
			
		} else {
			keysToDelete += entry.key
			logger.INFO ({'APPLICATION_REMOVED'}, entry.value.name);
		}
	}

	if (keysToDelete.size() > 0) {
		keysToDelete.each { key ->
			zdelete("/storage/applications#${key}")
		}
	}

    request.view = 'JSON'
    request.json.output = apps
    render()
}
 
 /**
 *
 * @success 200 Returns application details.
 * @error 404 If the application ID is not found.
 * @format application/json
 * @example
 * {
 *    "id": "1",
 *    "name": "my.app",
 *    "path": "path/to/parent"
 * }
 * 
 */
def onRetrieve() {
	def appid = request.params.applicationsId[]
	def storedData = storage.applications[appid]
	
    boolean qStatus=false;
	boolean isExport=false;
	String exportType=null;	
	
	//------- Check for /runnablestatus
    def pathInfo = event.pathInfo[];
    if (pathInfo != null &&
    	pathInfo.length() > 0 &&
    	pathInfo.equals('/runnablestatus'))
    {
        // System.out.println('\nPATHINFO: '+pathInfo);
    	qStatus=true
    }
    else
    {
    	// GET /applications/{id}?export&type=standalone  (or shared) 
    	String export = request.params.export[];
    	exportType='standalone';
    	if (export != null) {
    		isExport=true;
    		String type = request.params.type[];
    		if (type != null && type.length() > 0) {
    			if (type.equals('shared')) 
    				exportType = type;
    		}
    	}
    }
	
    
	def apphome = AppUtils.getApphome(appid)
	if (!apphome) {
		zdelete("/storage/applications#${appid}")
		request.status = 404
		return
	}

	// Request is either 
	//   A) an Export (Package) request
	//   B) an Application Status Check (nonRunnable Apps)
	//   or
	//   C) a Request for Application Details (default)
	
	//-------------------------------------- EXPORT
	if (isExport) {
		String revision=getRevision(apphome);
    	File exportDirectory = new File(apphome, 'export');
    	String fname;
        String fn;
    	if (exportType.equals('standalone')) {
    	  fn=apphome.name+'-all-'+revision;
    	} else {
    	  fn=apphome.name+'-'+revision;
    	}
        fname=fn+'.zip';
    	File exportFile = new File(exportDirectory, fname);
    	if (exportFile.exists()) {
    		exportFile.delete();
	    }
	    	  
        ArrayList clist = new ArrayList();
        clist.add('package');
        clist.add(exportType);
	    invoker = new CliTaskInvoker(apphome.canonicalPath, clist)
	    invoker.run()
	    dumpOutput(invoker)

	    if (!exportFile.exists())
	      {
	    	request.status = 500;
	    	return
	      }
	       
	    long fileSize=exportFile.length();
	    	
	    request.headers.out.'Content-type' = 'application/zip';
	    request.headers.out.'Content-Transfer-Encoding' = 'Binary';
	    request.headers.out.'Accept-Ranges' = 'bytes';
	    request.headers.out.'Content-Length' = fileSize;
	    request.headers.out.'Content-Disposition' = 'attachment; filename=\"'+fname+'\"';

	    int buffersize=32000;
	    byte[] buffer = new byte[buffersize];
	    FileInputStream  fin = null;
	    def out = request.outputStream[];
	    try
	      {
	      	fin = new FileInputStream (exportFile);
	      	int len=0;
	       	while ((len=fin.read(buffer,0,buffersize)) > 0)
	       	  {
	       		out.write(buffer,0,len);
	       	  }
	       	fin.close();
	       	fin=null;
	      }
	    catch (IOException e)
	      {
	      	if (fin != null)
	       		fin.close();
	    	request.status = 500;
	    	return
	      }
           
	    request.status = 200;
        return
	}
	else
	//-------------------------------------- App Status Check
	if (qStatus) {
	   def appData = storedData + getAppDetails(apphome) + [id:appid]; 
	   provideApplicationStatus(apphome, appData);
	   return
	}
	else
	//-------------------------------------- App Details
	{
	   request.view = 'JSON'
	   request.json.output = storedData + getAppDetails(apphome) + [id:appid]
	   render()
	   return
	}
}

/** 
 *
 * @success 201 Registered a Zero application (supports create {name,path}, 
 * open {name,path,existingApp=true}, and copy from repository {path,copyFrom}).
 * @error 400 Invalid request (message property in response object describes details).
 * @error 404 Open returns 404 if the specified application doesn't exist (must include config/ivy.xml).
 * @error 500 If the application already exists.
 * @format application/json
 * @example
 * To create a new application, post a JSON object with 'name' and 'path':
 * {
 *     "name": "my.zero.app",
 *     "path": "path/to/parent",
 *     "modulegroup": "experimental",
 *     "description": "application description"
 * }
 *
 * Result:  New application at path/to/parent/my.zero.app
 * 
 * 
 * To open an existing application, post a JSON object with 'name', 'path', and
 * 'existingApp=true':
 * {
 *     "name": "existing.app",
 *     "path": "path/to/parent",
 *     "existingApp": true
 * }
 * 
 * Result:  existing.app registered with the Application Builder
 *
 *
 * To create an application as a copy from the repository, post a JSON
 * object with 'copyFrom' (org:module[:version]) and 'path':
 * {
 *     "copyFrom": "zero:employee.demo",
 *     "path": "path/to/parent"
 * }
 *
 * Result:  Copy of employee.demo at path/to/parent/employee.demo
 */
def onCreate()
{    
    def entity = Json.decode(request.input[])
    
    if (!entity.path) {
    	request.status = 400
    	request.view = 'error'
    	request.error.message = MessageHandler.getMessage('PARAMETER_MISSING', '\'path\'');
    	render()
    	return
    }
    Boolean linkTo = entity.linkTo
    
    def approot = new File(entity.path)
    
    if (!approot.isAbsolute()) {
    	request.status = 400
    	request.view = 'error'
    	request.error.message = MessageHandler.getMessage('ABSOLUTE_REFERENCE_REQUIRED', '\'path\'');
    	render()
    	return
    }
    
    try {
    	if (entity.copyFrom) {//copy from Repo OR Linkto Repo module
    		copyLinktoApplicationFromRepository(entity,linkTo)
    		
    	} else if (entity.path && entity.name && entity.id) {
    		copyApplicationFromFileSystem(entity)
    		
    	} else if (entity.existingApp) {
	    	openApplication(entity)
	
    	} else {
    		newApplication(entity)
    	}
    
	} catch (IllegalStateException x) {
		// Thrown by CliTaskInvoker if ZERO_HOME env var is not set.
        request.status = 500
		request.view = 'error'
        request.error.message = MessageHandler.getMessage('ZERO_HOME_NOT_SET');
		render()
	}
}
 /** 
 *
 * @success 200 Updated a Zero application 
 * @error 400 Invalid request (message property in response object describes details).
 * @error 404 Open returns 404 if the specified application doesn't exist (must include config/ivy.xml).
 * @error 500 If application was started and could not be stopped during the update.
 * @format application/json
 * @example
 * To Update an application, post a JSON object with any or all of:
 * {
 *     "name": "my.zero.app",
 *     "path": "path/to/parent",
 *     "modulegroup": "experimental",
 *     "description": "application description"
 * }
 *
 * if description = changed  -> description updated
 * if modulegroup = changed  -> switch to new module group
 * if path        = changed  -> application moved to new path 
 * if name        = changed  -> application renamed 
 */
def onUpdate()
{    
	def appid = request.params.applicationsId[]
	def storedData = storage.applications[appid];
	def apphome
	def appRunning=false;

	def appInfo
	if (storedData) {
		apphome = new File(storedData.path, storedData.name)
		if (!apphome.exists()) {
		  request.status=404
		  request.view = 'error'
		  request.error.message = MessageHandler.getMessage('APPLICATION_NOT_FOUND',apphome.getCanonicalPath());
		  return
		}
		appInfo = storedData + getAppDetails(apphome);
		appRunning = appInfo.isRunning;
		if (appInfo.isRunning){
		  if (!AppUtils.stopApp(apphome)){
		    request.status = 500;
		    return;
		  }
		}
	} else {	
		request.status = 404
		return
	}
	
    def entity = Json.decode(request.input[])
    if (!entity.path) {
    	request.status = 400
    	request.view = 'error'
    	request.error.message = MessageHandler.getMessage('PARAMETER_MISSING', '\'path\'');
    	render()
    	return
    }
    
    def approot = new File(entity.path)    
    if (!approot.isAbsolute()) {
    	request.status = 400
    	request.view = 'error'
    	request.error.message = MessageHandler.getMessage('ABSOLUTE_REFERENCE_REQUIRED', '\'path\'');
    	render()
    	return
    }

    
    //----------------------------------------------------------------------------
    // entity:  JSON object w/ properties name, path, modulegroup, description
    //
    // applications.groovy:onUpdate()
    // - name change ==> move app directory; update ivy.xml
    // - path change ==> move app to new app root
    // - modulegroup ==> invoke "zero switch {modulegroup}" on the app
    // - description change ==> update ivy.xml
    //----------------------------------------------------------------------------

    try {
    	    boolean updates=false;
     	    if (!entity.description.equals(appInfo.description)) {
    	    	setDescription(apphome,entity.description);
                
    	    	updates=true;
    	    	
    	    	// System.out.println('Description Changed: '+appInfo.description+'=>'+entity.description+'\n');
     	    }
    	    if (!entity.modulegroup.equals(appInfo.modulegroup)) {
                ArrayList clist = new ArrayList();
                clist.add('switch');
                clist.add(entity.modulegroup);
    	    	invoker = new CliTaskInvoker(appInfo.path + File.separator + appInfo.name, clist)
    	    	invoker.run()
    	    	dumpOutput(invoker)
    	    	
    	    	updates=true;
    	    	// System.out.println('Modulegroup Changed: '+appInfo.modulegroup+'=>'+entity.modulegroup+' \n');
    	    }
    	    if (  (!entity.path.equals(appInfo.path)) ||
    	    	  (!entity.name.equals(appInfo.name))) {
    	    	if (!moveApplication(appInfo, entity.path, entity.name)) {
    	    		return;
    	    	}
    	    	updates=true;
    	    	// if (!entity.path.equals(appInfo.path))
    	    	//  	System.out.println('Path Changed: '+appInfo.path+'=>'+entity.path+' \n');
    	    	// if (!entity.name.equals(appInfo.name))
    	    	//  	System.out.println('Name Changed: '+appInfo.name+'=>'+entity.name+' \n');
    	    	  	    	  	    	
    	    	zpost('/storage#lock', '/applications')
    	    	zdelete("/storage/applications#${appid}");
    	    	if (!zcontains('/storage/applications')) {
    	    		storage.applications = [:]
    	    	}
    	    	zput("/storage/applications#${appid}", [name: entity.name, path: entity.path])
    	    	zpost('/storage#unlock', '/applications')

    	    	storedData = [id:appid, name:entity.name, path:entity.path];
    	    	apphome = new File(entity.path, entity.name);
    	    }

    	    if (updates)
    	      {
    	    	appInfo = getAppDetails(apphome) + storedData;
    	      }
    	    else
    	      {
    	    	// System.out.println('Nothing To Do\n'); 
    	      }	

   	    	request.view = 'JSON'
    	    request.json.output = appInfo; 
    		request.headers.out.Location = getAbsoluteUri("/resources/applications/${storedData.id}")
    		request.status = 200
     		render()
 
    } catch (IllegalStateException x) {
		// Thrown by CliTaskInvoker if ZERO_HOME env var is not set.
        request.status = 500
		request.view = 'error'
        request.error.message = MessageHandler.getMessage('ZERO_HOME_NOT_SET');
		render()
	}
}
 
 /** 
 *
 * @success 204 Application was deleted (metadata only if URI includes query parameter ?metadata).
 * @error 404 If the application ID was not found.
 * @error 500 If the application was started and could not be stopped.
 * 
 */
def onDelete() {
	def appid = request.params.applicationsId[]
	def storedData = storage.applications[appid]
	
	if (storedData) {
		def apphome = new File(storedData.path, storedData.name)
	
		if (apphome.exists()) {
			if (!AppUtils.stopApp(apphome)){
				request.status = 500;
				return;
			}

			if (!request.params.metadata[]) {
				FileUtils.delete(apphome)
			}
		
			zdelete("/storage/applications#${appid}")
			
			request.status = 204
			return
		}
	}

	request.status = 404
}
 
 
def onPostMember() {
	def appid = request.params.applicationsId[]
	def storedData = storage.applications[appid]
	
	if (storedData) {
    	def apphome = new File(storedData.path, storedData.name)
		
	    def entity = Json.decode(request.input[])
	    if (entity?.command == 'update') {
	    	CliTaskInvoker invoker = new CliTaskInvoker(apphome.canonicalPath, 'update')
	    	rc = invoker.run()
	    	dumpOutput(invoker)
		
	    	if (rc != 0) {
	    		request.status = 500
	    		request.view = 'error'

	    		// CWPZT0601E is the message id for the generic 'fail' message
	    		// that advises users to re-run a command with -logfile.  Strip
	    		// that off for AB usage.
	    		request.error.message = (invoker.stdout + invoker.stderr).findAll {
	    			!it.startsWith('CWPZT0601E')
	    		}
	    		render()
	    		return
	    	}
	    	
	    	request.status = 204
	    }
	    
	    else {
	    	// Unknown command
	    	request.status = 400
	    }
	    
	} else {
		// appid not found
		request.status = 404
	}
}

 
def copyApplicationFromFileSystem(entity)
{    
	def RENAME_TARGET_IF_EXISTS=false;
	def id = entity.id;
	def storedData = storage.applications[id];
	def invoker
	def rc=0

	def sourceInfo
	if (storedData) {
		def apphome = new File(storedData.path, storedData.name)
		if (!apphome.exists()) {
		  request.status=404
		  return
		}
		sourceInfo = storedData + getAppDetails(apphome);
	} else {	
		request.status = 404
		return
	}
	
    if (!entity.path) {
    	request.status = 400
    	request.view = 'error'
    	request.error.message = MessageHandler.getMessage('PARAMETER_MISSING', '\'path\'');
    	render()
    	return
    }
    
    //----------------------------------------------------------------------------
    // entity:  JSON object w/ properties name, path, modulegroup, description
    //----------------------------------------------------------------------------
    try {
        new File(entity.path).mkdirs()
        
        File sourcePath = new File(storedData.path, storedData.name);
        File targetPath = new File(entity.path, entity.name);
     
        // System.out.println('Source:  '+sourcePath.getCanonicalPath());
        // System.out.println('Target:  '+targetPath.getCanonicalPath());
        
        if (!sourcePath.exists())
          {
            request.status = 404
    		request.view = 'error'
    		request.error.message = MessageHandler.getMessage('APPLICATION_NOT_FOUND', sourcePath.getCanonicalPath());
            render()
            return
          }
        
        if (RENAME_TARGET_IF_EXISTS) {
        	/*---- Do we construct a Name if copy target already exists ---*/
        	if (targetPath.exists()) {
        		entity.name = 'CopyOf'+storedData.name;
        		targetPath = new File(entity.path, entity.name);
                // System.out.println('  Target Exists - trying ... '+targetPath.getCanonicalPath());
        	}
        	int retries=3;
        	while (targetPath.exists()) {
        		retries--;
        		if (retries==0) break;
        		
        		entity.name = 'CopyOf'+entity.name;
        		targetPath = new File(entity.path, entity.name);
                // System.out.println('  Target Exists - trying ... '+targetPath.getCanonicalPath());             
        	}
        }
        
        if (targetPath.exists())
          {
            request.status = 406
    		request.view = 'error'
    		request.error.message = MessageHandler.getMessage('APPLICATION_EXISTS', targetPath.getCanonicalPath());
    		render()
            return
          }

        try {
           FileExclusion.exclusions = [
             new File(sourcePath, 'logs'),
             new File(sourcePath, '.zero/private'),
           ];
           copyDirectory(sourcePath, targetPath);
        }
        catch(IOException e) {
           request.status = 500
           request.view = 'error'
           request.error.message = MessageHandler.getMessage('APPLICATION_COPY_FAILED', sourcePath.getCanonicalPath(), targetPath.getCanonicalPath());
           render()
           return
        }
        
 	    if (!entity.description.equals(sourceInfo.description)) {
	    	setDescription(targetPath,entity.description);
 	    }
 	    if (!entity.name.equals(sourceInfo.name)) {
 	        setModuleName(targetPath);
 	    }
	    if (!entity.modulegroup.equals(sourceInfo.modulegroup)) {
            ArrayList clist = new ArrayList();
            clist.add('switch');
            clist.add(entity.modulegroup);
	    	invoker = new CliTaskInvoker(targetPath.canonicalPath, clist)
	    	rc=invoker.run()
	    	dumpOutput(invoker)
	    	
	    	if (rc != 0) {
	    		request.view = 'error'
	    		request.status = 406
	    		request.error.message = MessageHandler.getMessage('MODULEGROUP_SWITCH_FAILURE', entity.modulegroup);
	    		render()
	    		return
	    	}
	    }
        // Put Application in the Applications List regardless of whether or
        // not it resolves correctly
    	storedData = saveApplication(entity.path, entity.name);
        
    	invoker = new CliTaskInvoker(targetPath.canonicalPath, 'resolve')
    	rc = invoker.run()
    	// dumpOutput(invoker)
    	
    	// Return Resolve Output Messages in the 'notRunnableDetail' if the 
    	// Application fails to Resolve
    	def cliOutput;
		if (rc != 0) {
			// CWPZT0601E is the message id for the generic 'fail' message
			// that advises users to re-run a command with -logfile.  Strip
			// that off for AB usage.
			cliOutput = (invoker.stdout + invoker.stderr).findAll {
				!it.startsWith('CWPZT0601E')
			}

		} else {
			// resolve won't catch problems with config
			// if resolve succeeded, then try 'zero version' to check config
			// note that if resolve failed, then there is no resolved.properties
			//  and there is no point in running 'zero version' since it would
			//  be run at the CLI level
			invoker = new CliTaskInvoker(targetPath.canonicalPath, 'version')
			rc = invoker.run()
			// dumpOutput(invoker)
			if (rc==0)
				cliOutput=[];
			else
			    {
				// CWPZT0601E is the message id for the generic 'fail' message
				// that advises users to re-run a command with -logfile.  Strip
				// that off for AB usage.
				cliOutput = (invoker.stdout + invoker.stderr).findAll {
					!it.startsWith('CWPZT0601E')
				}
			}
		}

		request.status = 201
		request.view = 'JSON'
		request.json.output = ['app' : getAppDetails(targetPath) + storedData, 'notRunnableDetail' : cliOutput]
		request.headers.out.Location = getAbsoluteUri("/resources/applications/${storedData.id}")
		render()             
        return 
    } catch (IllegalStateException x) {
		// Thrown by CliTaskInvoker if ZERO_HOME env var is not set.
        request.status = 500
		request.view = 'error'
        request.error.message = MessageHandler.getMessage('ZERO_HOME_NOT_SET');
		render()
		return 
	}
}

 
def copyLinktoApplicationFromRepository(entity,linkTo)
{
	if (!entity.name)
	{
		// For now, we'll require copyFrom to include org and module; version is optional
		def segmentsOfCopyFrom = entity.copyFrom.split(':')
		if ((segmentsOfCopyFrom.size() < 2) || (segmentsOfCopyFrom.size() > 3)
				|| (segmentsOfCopyFrom[0].size() == 0) || (segmentsOfCopyFrom[1].size() == 0)) {
			request.status = 400
			request.view = 'error'
		    request.error.message = MessageHandler.getMessage('INVALID_MODULE_REFERENCE',"${entity.copyFrom}");
			render()
			return
		}
	
		entity.name = segmentsOfCopyFrom[1]
	}
	
    new File(entity.path).mkdirs()
    
	def apphome = new File(entity.path, entity.name)
    if (apphome.exists())
    {
        request.status = 500
		request.view = 'error'
    	request.error.message = MessageHandler.getMessage('APPLICATION_EXISTS', "${apphome}");
		render()
        return
    }
    
    /*
     * See the two step create comment in newApplication().  This case is similar.
     */
    def rc
    def path = apphome.getCanonicalPath();
    List<String> cmd = new ArrayList<String>();
	cmd.add('create');
	cmd.add(path);
	//CLI - copy from repo or Linkto module from repo
	linkTo ? cmd.add('linkto') : cmd.add('from');
	cmd.add(entity.copyFrom);
    ModuleGroupTaskInvoker mginvoker = new ModuleGroupTaskInvoker(entity.modulegroup, cmd)
	rc = mginvoker.run()
	dumpOutput(mginvoker)
	
	if (rc != 0) {
		if (apphome.exists()) {
			FileUtils.delete(apphome)
		}
		request.view = 'error'
		request.status = 500
		
        // CWPZT0601E is the message id for the generic 'fail' message
        // that advises users to re-run a command with -logfile.  Strip
        // that off for AB usage.
        request.error.message = (mginvoker.stdout + mginvoker.stderr).findAll {
        	!it.startsWith('CWPZT0601E')
        }
       
		render()
		return
	}
	
	CliTaskInvoker cli = new CliTaskInvoker(path, ['switch', entity.modulegroup]);
	rc = cli.run()
	dumpOutput(cli);
	
	if (rc != 0) {
		if (apphome.exists()) {
			FileUtils.delete(apphome)
		}
		request.status = 500
		request.view = 'error'
		
        // CWPZT0601E is the message id for the generic 'fail' message
        // that advises users to re-run a command with -logfile.  Strip
        // that off for AB usage.
        request.error.message = (cli.stdout + cli.stderr).findAll {
        	!it.startsWith('CWPZT0601E')
        }
		render()
		return
	}
	
    def storedData = saveApplication(entity.path, entity.name)
 
    request.view = 'JSON'
    request.json.output = getAppDetails(apphome) + storedData
    request.headers.out.Location = getAbsoluteUri("/resources/applications/${storedData.id}")
    request.status = 201
    render()
}

def moveApplication(appInfo, targetPath, targetName)
{
	def currentAppHome = new File(appInfo.path, appInfo.name);
	def newAppHome     = new File(targetPath, targetName);
	
    if (newAppHome.exists())
    {
        request.status = 500
		request.view = 'error'
	    request.error.message = MessageHandler.getMessage('APPLICATION_EXISTS', "${newAppHome}");
		render()
        return false
    }
    
    // Move File
    // Covers all of the Move/Rename semantics
    //  - Rename Application File name in its current directory 
    //  - Move   Application with its current name to new directory
    //  - Move   Application with its new Application File Name to new directory

    new File(targetPath).mkdirs();

    boolean success = currentAppHome.renameTo(newAppHome);
    if (!success) {
        request.status = 500
		request.view = 'error'
	    request.error.message = MessageHandler.getMessage('APPLICATION_MOVE_FAILED', "${currentAppHome}","${newAppHome}");
		render()
        return false
    }
    	
	if (!appInfo.name.equals(targetName)) {
	  setModuleName(newAppHome);
	}
	
	def invoker = new CliTaskInvoker(newAppHome.canonicalPath, 'resolve')
	def rc = invoker.run()
	// dumpOutput(invoker)
	
	if (rc != 0) {
		request.view = 'error'
		request.status = 406
	    request.error.message = MessageHandler.getMessage('APPLICATION_CREATE_FAILURE', "${newAppHome}");
		render()
		return false
	}

    return true  
}

def newApplication(entity) {
	def apphome = new File(entity.path, entity.name)
    if (apphome.exists())
    {
        request.status = 500
	    request.error.message = MessageHandler.getMessage('APPLICATION_EXISTS', "${apphome}");
		request.view = 'error'
		render()
        return
    }	
	
	new File(entity.path).mkdirs()
    
	def path = new File(entity.path, entity.name)
	List<String> cmd = new ArrayList<String>();
	cmd.add('create');
	cmd.add(path.absolutePath);
	ModuleGroupTaskInvoker invoker = new ModuleGroupTaskInvoker(entity.modulegroup, cmd)
	int rc = invoker.run()
	dumpOutput(invoker)
	
	if (rc != 0) {
		if (apphome.exists()) {
			FileUtils.delete(apphome)
		}
        // CWPZT0601E is the message id for the generic 'fail' message
		// that advises users to re-run a command with -logfile.  Strip
		// that off for AB usage.
		def cliOutput = (invoker.stdout + invoker.stderr).findAll {
			!it.startsWith('CWPZT0601E')
		}

        request.status = 500
		request.view = 'error'
	    request.error.message = cliOutput //this is a better message; explains why CLI failed
	    //request.error.message = MessageHandler.getMessage('APPLICATION_CREATE_FAILURE', "${path}");
		render()
		return
	}
	
	def storedData = saveApplication(entity.path, entity.name)
	
	setDescription(apphome, entity.description)

	request.view = 'JSON'
    request.json.output = getAppDetails(apphome) + storedData
	request.headers.out.Location = getAbsoluteUri("/resources/applications/${storedData.id}")
	request.status = 201
	render()
}

def openApplication(entity) {
	def apphome = new File(entity.path, entity.name)
	def ivyFile = new File(apphome, 'config/ivy.xml')
	if (!ivyFile.exists()) {
		request.status = 404
		request.view = 'error'
		request.error.message = MessageHandler.getMessage('APPLICATION_NOT_FOUND', "${apphome}");
		render()
		return
	}
	
	// If the app is already registered (in storage zone), then return 303 (SEE OTHER) with
	// a pointer to the existing resource.
	def apps = storage.applications[]
	def match = apps.find { (it.value.name == entity.name) && (it.value.path == entity.path)}
	
	if (match) {
		request.status = 303
		request.headers.out.Location = getAbsoluteUri("/resources/applications/${match.key}")
		
	} else {
		def storedData = saveApplication(entity.path, entity.name)
		
		CliTaskInvoker invoker = new CliTaskInvoker(apphome.canonicalPath, 'resolve')
		rc = invoker.run()
		// dumpOutput(invoker)
		
		def cliOutput;
		if (rc != 0) {
			// CWPZT0601E is the message id for the generic 'fail' message
			// that advises users to re-run a command with -logfile.  Strip
			// that off for AB usage.
			cliOutput = (invoker.stdout + invoker.stderr).findAll {
				!it.startsWith('CWPZT0601E')
			}

		} else {
			// resolve won't catch problems with config
			// if resolve succeeded, then try 'zero version' to check config
			// note that if resolve failed, then there is no resolved.properties
			//  and there is no point in running 'zero version' since it would
			//  be run at the CLI level
			invoker = new CliTaskInvoker(apphome.canonicalPath, 'version')
			rc = invoker.run()
			// dumpOutput(invoker)
			if (rc==0)
				cliOutput=[];
			else
			    {
				// CWPZT0601E is the message id for the generic 'fail' message
				// that advises users to re-run a command with -logfile.  Strip
				// that off for AB usage.
				cliOutput = (invoker.stdout + invoker.stderr).findAll {
					!it.startsWith('CWPZT0601E')
				}
			}
		}
		
		request.status = 201
		request.view = 'JSON'
		request.json.output = ['app' : getAppDetails(apphome) + storedData, 'notRunnableDetail' : cliOutput]
		request.headers.out.Location = getAbsoluteUri("/resources/applications/${storedData.id}")
		render()
	}
}

/**
 * apphome is a File pointing to the application home.
 */
def getAppDetails(apphome) {
	if (!apphome.exists()) {
		return null
	}

	def details = [
		description: getDescription(apphome),
		modulegroup: getModulegroup(apphome),
		rootUri: getRootUri(apphome),
		isRunning: AppUtils.isRunning(apphome),
		isRunnable: AppUtils.isRunnable(apphome),
		canDebug: checkDebug(apphome)
	]

	return details
}

def dumpOutput(invoker) {
	invoker.getStderr().each {
		System.err.println it
	}
	invoker.getStdout().each {
		System.out.println it
	}
}

/**
 * @return Map of stored data (id, name, path).
 */
def saveApplication(path, name) {
	zpost('/storage#lock', '/applications/nextid')
	
	// id is a string based on an auto-incrementing id.
	// Ensures that deleted member ids are not reused.
	def counter = zget('/storage/applications/nextid', 0)
	def id = '' + counter
	storage.applications.nextid = counter+1
	
	if (!zcontains('/storage/applications')) {
		storage.applications = [:]
	}
	
	zput("/storage/applications#${id}", [name: name, path: path])
	
	zpost('/storage#unlock', '/applications/nextid')
	
	return [id:id, name:name, path:path]
}

// Returns a groovy.util.Node on success or an error String on failure.
def parseIvyFile(appHome) {
	def ivyFile = new File(appHome, 'config/ivy.xml');
	String error = 'Failed to parse ' + ivyFile.getAbsolutePath() + ': ';
	if (ivyFile.exists()) {
		try {
			try {
				def node = new XmlParser().parse(ivyFile);
				return node;
			} catch (ParserConfigurationException e) {
				error += e.getMessage();
			} catch (SAXParseException e) {
				error += e.getMessage();
			} catch (IOException e) {
				error += e.getMessage();
			}
		}
	}
	else {
		error += 'File does not exist.';
	}
	return error;
}

def writeIvyFile(appHome, docRootNode) {
	def ivyFile = new File(appHome, 'config/ivy.xml');
	try {
	    def writer = new UTF8FileWriter(ivyFile);
		new XmlNodePrinter(new PrintWriter(writer)).print(docRootNode);
		writer.close();
	} catch (IOException e) {
		logger.WARNING ({'WRITE_FAILURE'}, ivyFile.getAbsolutePath(),e);
	}
}

def getRevision(appHome) {
	def result = parseIvyFile(appHome);
	if (result instanceof Node) {
		def attr = result.info[0].attributes();
		return (String) attr.get('revision');
	}
	return result.toString();
}

def getDescription(appHome) {
	def result = parseIvyFile(appHome);
	if (result instanceof Node) {
		return result.info[0].description[0].text();
	}
	return result.toString();
}

// Save the description to ivy.xml.  Fails if the file is invalid.
def setDescription(appHome, description) {
	def result = parseIvyFile(appHome);
	if (result instanceof Node) {
		result.info[0].description[0].setValue(XMLEncoder.escapeXML(description));
		writeIvyFile(appHome,result);
	}
	return result.toString();
}

//Save the module name to ivy.xml.  Fails if the file is invalid.
def setModuleName(appHome) {
	def result = parseIvyFile(appHome);
	if (result instanceof Node) {
		def attr = result.info[0].attributes();
		attr.put('module',XMLEncoder.escapeXML(appHome.name));
 		attr = result.publications[0].artifact[0].attributes();
		attr.put('name',XMLEncoder.escapeXML(appHome.name));
		writeIvyFile(appHome,result);
	}
	return result.toString();
}

def getModulegroup(appHome) {
	def resolveHistory = new File(appHome, '.zero/shared/resolve.history')
	if (resolveHistory.exists()) {
		def history = new java.util.Properties()
		def fis = new FileInputStream(resolveHistory)
		try {
			history.load(fis)
		} finally {
			fis.close()
		}
		return history.modulegroup
	}
	
	return null
}

def getRootUri(apphome) {
	def rootUri = null
	
	// Load the application's config
	try {
		def context = new zero.core.context.Context()
		zero.core.config.ConfigLoader.initialize(apphome.getAbsolutePath(), context)
    	
		def contextRoot = context.zget('/config/contextRoot')

		def prefix = context.zget('/config/externalUriPrefix', null)
		if (prefix) {
			rootUri = "$prefix$contextRoot"
		
		} else {
			// Server name is the same as that used to reach the IDE
			def server = zget('/request/serverName')
    	
			def port = context.zget('/config/http/port')
    	
			if (port > 0) {
				rootUri = "http://$server:$port$contextRoot"
        	
			} else {
				port = context.zget('/config/https/port')
				if (port > 0) {
					rootUri = "https://$server:$port$contextRoot"
				}
			}
		}
			
	} catch (Exception x) {
		logger.WARNING ({'CONFIG_LOAD_FAILURE'}, apphome.getAbsolutePath(), x);
	}
	
	return rootUri
}

def copyDirectory(File source, File target) throws IOException {
   if (source.isDirectory()) {
       if (!target.exists()) {
           target.mkdir();
       }
        
       String[] children = source.list();
       for (int i=0; i<children.length; i++) {
    	  File src=new File(source,children[i]);
    	  String srcPath=src.getCanonicalPath();
    	  boolean copy=true;
     	  for (int j=0; j<FileExclusion.exclusions.length; j++) {
     		  if (srcPath.equals(FileExclusion.exclusions[j])) {
     			  copy=false;
     			  break;
     		  }
     	  }
    	  if (copy)
    	  	{
    		  File tgt=new File(target,children[i])
    		  copyDirectory(src, tgt);
    	  	}
       }
   } else {
       InputStream input = new FileInputStream(source);
       OutputStream output = new FileOutputStream(target);

       byte[] buf = new byte[4096];
       int len;
       while ((len = input.read(buf)) > 0) {
           output.write(buf, 0, len);
       }
       input.close();
       output.close();
   }
}

def deleteDirectory(File dir) {
    if (!dir.isDirectory()) return false;
    
     // Delete all files and subdirectories one by one
    File[] files = dir.listFiles();
    if (files != null) {
        for (int i = 0; i < files.length; i++) {
            File file = files[i];
            if (file.isDirectory())
               deleteDirectory(file);
            else
               file.delete();
        }
    }

    // Director should be empty .. delete it
    return dir.delete();  
}

def provideApplicationStatus(apphome, appData) { 
	if (appData.isRunnable) {
	  // If application is runnable - simply indicate so along with the app details
	  // This could occur if a problem had occurred that was rectified before the
	  // 'runnable' Application status was updated.  The client should simply
	  // indicate that the application is runnabe and update the AppManager runnable
	  // status
	  request.view = 'JSON'
	  request.json.output = ['app' : appData, 'notRunnableDetail' : []]
	  render()
	  return
	}
	
	def result = parseIvyFile(apphome);
	if (!(result instanceof Node)) {
		String[] message = new String[1];
		message[0] = result.toString();
		request.status = 200
		request.view = 'JSON'
		request.json.output = ['app' : appData, 'notRunnableDetail' : message]
		render()  
		return
	}
	
	// Application is not Runnable ... lets investigate and see if we can tell why
	
	def invoker = new CliTaskInvoker(apphome.canonicalPath, 'resolve')
	int rc = invoker.run()
	// dumpOutput(invoker)
    	
    	// Return Resolve Output Messages in the 'notRunnableDetail' if the 
    	// Application fails to Resolve
    def cliOutput;
	if (rc != 0) {
		// CWPZT0601E is the message id for the generic 'fail' message
		// that advises users to re-run a command with -logfile.  Strip
		// that off for AB usage.
		
		String eMessage = MessageHandler.getMessage('APPLICATION_RESOLVE_FAILURE', apphome.canonicalPath);
		cliOutput = [eMessage+'\n'] + (invoker.stdout + invoker.stderr).findAll {
			!it.startsWith('CWPZT0601E')
		}
	} else {
		// resolve won't catch problems with config
		// if resolve succeeded, then try 'zero version' to check config
		// note that if resolve failed, then there is no resolved.properties
		//  and there is no point in running 'zero version' since it would
		//  be run at the CLI level
		invoker = new CliTaskInvoker(apphome.canonicalPath, 'version')
		rc = invoker.run()
		// dumpOutput(invoker)
		if (rc==0)
			cliOutput=[];
		else
		{
			// CWPZT0601E is the message id for the generic 'fail' message
			// that advises users to re-run a command with -logfile.  Strip
			// that off for AB usage.
			cliOutput = (invoker.stdout + invoker.stderr).findAll {
				!it.startsWith('CWPZT0601E')
			}
		}
	}

	request.status = 200
	request.view = 'JSON'
	request.json.output = ['app' : appData, 'notRunnableDetail' : cliOutput]
	render()             
	return
}

def checkDebug(apphome) {	
    if(zget("/config/debugEnabled")==false) {
    	return false
    }
    def debugEnabled = false
	try {
		// check for JDI classes
	    Class.forName("com.sun.jdi.VirtualMachine")
	    // now check for ZSO 1.1 or greater
	    def deps = AppUtils.getTransitiveDependenciesAsList(apphome, apphome.name)
	    for(Map<String,String> d: deps) { 
	    	if(d.name=="zero.management.zso") {
	    		def t = d.rev
	    		def rev = d.rev.split("\\.")
	    		def major = Integer.parseInt(rev[0]);
	    		def minor = Integer.parseInt(rev[1]);
	    		if(major > 1 || (major==1 && minor>=1)) {
	    			debugEnabled = true
	    		}
	    		break
	    	}
	    }
	} catch(ClassNotFoundException e) { //debugEnabled=false
	}	             
	return debugEnabled
}
