/*
 * ============================================================================
 * Licensed Materials - Property of IBM
 * Project  Zero
 *
 * (C) Copyright IBM Corp. 2007  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 * ============================================================================
 */


// 
// File: /app/resources/processes.groovy
// 
// Author: Dan Jemiolo (danjemiolo)
// 
// This script provides the REST API for starting, stopping, and monitoring 
// Zero applications via the IDE. The "start" task is delegated to the 
// Zero CLI; "stop" and "status" are handled via the zero.management REST APIs.
//

import java.net.HttpURLConnection
import java.net.URL
import zero.ide.services.CliTaskInvoker
import zero.json.Json
import zero.ide.services.AppUtils
import zero.management.appconfig.Application
import zero.core.events.EventEngine;
import zero.appbuilder.resources.MessageHandler

/**
 *
 * @success 200 If the application is currently running.
 * @error 404 If the application is not running.
 *
 */
def onRetrieve()
{
	def appid = request.params.processesId[]
	def apphome = AppUtils.getApphome(appid)
	if (apphome && AppUtils.isRunning(apphome)) {
		request.status = 200
	} else {
		request.status = 404
	}
}

/**
 *
 * @success 201 The application was started successfully. It can be reached using the port specified in zero.config.
 * @success 204 If the application is already running.
 * @error 404 If the application does not exist.
 * @error 500 If the application has unresolved dependencies or missing files.
 *
 */
def onCreate()
{
    def payload = Json.decode(request.input[])
    def apphome = AppUtils.getApphome(payload.appid)
    
    if (!apphome) {
    	request.status = 404
    	return
    }
    
    //
    // app is already running - do nothing
    //
    if (AppUtils.isRunning(apphome)) {
        request.status = 204
    	return
    }
    
    def args = ["start"]
    if (null != payload) {
        if (payload.containsKey("debug") && payload.debug) {
        	startDebug(payload.appid,args);
        }
    }
    
	invoker = new CliTaskInvoker(apphome.canonicalPath, args)
	def rc = invoker.run()
    dumpOutput(invoker)
    
    if (rc != 0) {
    	// stop debugger 
    	
        request.status = 409
        request.view = 'error'
        
        // CWPZT0601E is the message id for the generic "fail" message
        // that advises users to re-run a command with -logfile.  Strip
        // that off for AB usage.
        request.error.message = (invoker.stdout + invoker.stderr).findAll {
        	!it.startsWith("CWPZT0601E")
        }

		render()
        return
    }
	
    request.headers.out.Location = getAbsoluteUri(request.path[] + "/" + payload.appid)
    request.status = 201
}

/**
 *
 * @success 204 The application was stopped successfully. If the application was not running, no action was taken.
 * @error 404 If there is no application with the given ID.
 * @error 500 If the application could not be stopped.
 *
 */
def onDelete()
{
   	def appid = request.params.processesId[]
	def apphome = AppUtils.getApphome(appid)
	if (!apphome) {
		request.status = 404
		return
	}   	
    request.status = AppUtils.stopApp(apphome) ? 204 : 500;
    stopDebug(appid);    
}

/*
 * This method starts a thread listening for a debug connection from a 
 * VM that is to be debugged.  The VM will connect when it is started by 
 * the 'zero start' command
 */
def startDebug(def appId, def args) {
	Map<String, Object> eventData = new HashMap<String, Object>();
	eventData.put("appId",appId);
	eventData.put("args",args);
	EventEngine.fire("startDebug",eventData);
}		

def stopDebug(def appId) {
	Map<String, Object> eventData = new HashMap<String, Object>();
	eventData.put("appId",appId);
	EventEngine.fire("stopDebug",eventData);
}

def dumpOutput(invoker) {
	invoker.getStderr().each {
		System.err.println it
	}
	invoker.getStdout().each {
		System.out.println it
	}
}